@@ -1,22 +1,29 @@
 #![allow(clippy::too_many_arguments)]
+mod completer;
 mod util;
 
-use std::time::{Duration, Instant};
+use std::{
+    io::IsTerminal,
+    time::{Duration, Instant},
+};
 
-use arboard::Clipboard;
 use async_std::sync::Arc;
 use clap::{Args, CommandFactory, Parser, Subcommand};
 use color_eyre::{eyre, eyre::Context};
+use completer::enter_code;
 use console::{style, Term};
 use futures::{future::Either, Future, FutureExt};
 use indicatif::{MultiProgress, ProgressBar};
-use std::{io::Write, path::PathBuf};
-
 use magic_wormhole::{
     forwarding, transfer,
     transit::{self, TransitInfo},
-    MailboxConnection, Wormhole,
+    MailboxConnection, ParseCodeError, ParsePasswordError, Wormhole,
 };
+use std::{io::Write, path::PathBuf};
+use tracing_subscriber::EnvFilter;
+
+#[cfg(feature = "clipboard")]
+use arboard::Clipboard;
 
 fn install_ctrlc_handler(
 ) -> eyre::Result<impl Fn() -> futures::future::BoxFuture<'static, ()> + Clone> {
@@ -31,11 +38,11 @@ fn install_ctrlc_handler(
             let mut has_notified = notifier2.0.lock().await;
             if *has_notified {
                 /* Second signal. Exit */
-                log::debug!("Exit.");
+                tracing::debug!("Exit.");
                 std::process::exit(130);
             }
             /* First signal. */
-            log::info!("Got Ctrl-C event. Press again to exit immediately");
+            tracing::info!("Got Ctrl-C event. Press again to exit immediately");
             *has_notified = true;
             notifier2.1.notify_all();
         })
@@ -133,7 +140,7 @@ enum ForwardCommand {
     )]
     Serve {
         /// List of ports to open up. You can optionally specify a domain/address to forward remote ports
-        #[arg(value_name = "[DOMAIN:]PORT", action = clap::ArgAction::Append, value_hint = clap::ValueHint::Hostname)]
+        #[arg(value_name = "[DOMAIN:]PORT", required = true, action = clap::ArgAction::Append, value_hint = clap::ValueHint::Hostname)]
         targets: Vec<String>,
         #[command(flatten)]
         common: CommonArgs,
@@ -263,27 +270,21 @@ async fn main() -> eyre::Result<()> {
     let mut term = Term::stdout();
 
     if app.log {
-        env_logger::builder()
-            .filter_level(log::LevelFilter::Debug)
-            .filter_module("magic_wormhole::core", log::LevelFilter::Trace)
-            .filter_module("mio", log::LevelFilter::Debug)
-            .filter_module("ws", log::LevelFilter::Error)
-            .try_init()?;
-        log::debug!("Logging enabled.");
+        tracing_subscriber::fmt()
+            .with_max_level(tracing::Level::TRACE)
+            .with_env_filter(EnvFilter::new(
+                "magic_wormhole::core=trace,mio=debug,ws=error",
+            ))
+            .with_target(false)
+            .init();
+        tracing::trace!("Logging enabled.");
     } else {
-        env_logger::builder()
-            .filter_level(log::LevelFilter::Info)
-            .filter_module("ws", log::LevelFilter::Error)
-            .format_timestamp(None)
-            .format_target(false)
-            .try_init()?;
-    }
-
-    let mut clipboard = Clipboard::new()
-        .map_err(|err| {
-            log::warn!("Failed to initialize clipboard support: {}", err);
-        })
-        .ok();
+        tracing_subscriber::fmt()
+            .with_max_level(tracing::Level::INFO)
+            .with_env_filter(EnvFilter::new("mio=debug"))
+            .with_target(false)
+            .init();
+    };
 
     match app.command {
         WormholeCommand::Send {
@@ -304,7 +305,6 @@ async fn main() -> eyre::Result<()> {
                     true,
                     transfer::APP_CONFIG,
                     Some(&sender_print_code),
-                    clipboard.as_mut(),
                 )),
                 ctrl_c(),
             )
@@ -342,7 +342,6 @@ async fn main() -> eyre::Result<()> {
                     true,
                     transfer::APP_CONFIG,
                     Some(&sender_print_code),
-                    clipboard.as_mut(),
                 ));
                 match futures::future::select(connect_fut, ctrl_c()).await {
                     Either::Left((result, _)) => result?,
@@ -382,7 +381,6 @@ async fn main() -> eyre::Result<()> {
                     false,
                     transfer::APP_CONFIG,
                     None,
-                    clipboard.as_mut(),
                 ));
                 match futures::future::select(connect_fut, ctrl_c()).await {
                     Either::Left((result, _)) => result?,
@@ -407,7 +405,7 @@ async fn main() -> eyre::Result<()> {
             ..
         }) => {
             // TODO make fancy
-            log::warn!("This is an unstable feature. Make sure that your peer is running the exact same version of the program as you. Also, please report all bugs and crashes.");
+            tracing::warn!("This is an unstable feature. Make sure that your peer is running the exact same version of the program as you. Also, please report all bugs and crashes.");
             /* Map the CLI argument to Strings. Use the occasion to inspect them and fail early on malformed input. */
             let targets = targets
                 .into_iter()
@@ -456,7 +454,6 @@ async fn main() -> eyre::Result<()> {
                     true,
                     app_config,
                     Some(&server_print_code),
-                    clipboard.as_mut(),
                 ));
                 let (wormhole, _code, relay_hints) =
                     match futures::future::select(connect_fut, ctrl_c()).await {
@@ -481,20 +478,11 @@ async fn main() -> eyre::Result<()> {
             ..
         }) => {
             // TODO make fancy
-            log::warn!("This is an unstable feature. Make sure that your peer is running the exact same version of the program as you. Also, please report all bugs and crashes.");
+            tracing::warn!("This is an unstable feature. Make sure that your peer is running the exact same version of the program as you. Also, please report all bugs and crashes.");
             let mut app_config = forwarding::APP_CONFIG;
             app_config.app_version.transit_abilities = parse_transit_args(&common);
-            let (wormhole, _code, relay_hints) = parse_and_connect(
-                &mut term,
-                common,
-                code,
-                None,
-                false,
-                app_config,
-                None,
-                clipboard.as_mut(),
-            )
-            .await?;
+            let (wormhole, _code, relay_hints) =
+                parse_and_connect(&mut term, common, code, None, false, app_config, None).await?;
 
             let offer = forwarding::connect(
                 wormhole,
@@ -504,10 +492,10 @@ async fn main() -> eyre::Result<()> {
                 &ports,
             )
             .await?;
-            log::info!("Mapping the following open ports to targets:");
-            log::info!("  local port -> remote target (no address = localhost on remote)");
+            tracing::info!("Mapping the following open ports to targets:");
+            tracing::info!("  local port -> remote target (no address = localhost on remote)");
             for (port, target) in &offer.mapping {
-                log::info!("  {} -> {}", port, target);
+                tracing::info!("  {} -> {}", port, target);
             }
             if noconfirm || util::ask_user("Accept forwarded ports?", true).await {
                 offer.accept(ctrl_c()).await?;
@@ -573,12 +561,11 @@ type PrintCodeFn = dyn Fn(&mut Term, &magic_wormhole::Code, &Option<url::Url>) -
 async fn parse_and_connect(
     term: &mut Term,
     common_args: CommonArgs,
-    code: Option<String>,
+    mut code: Option<String>,
     code_length: Option<usize>,
     is_send: bool,
     mut app_config: magic_wormhole::AppConfig<impl serde::Serialize + Send + Sync + 'static>,
     print_code: Option<&PrintCodeFn>,
-    clipboard: Option<&mut Clipboard>,
 ) -> eyre::Result<(Wormhole, magic_wormhole::Code, Vec<transit::RelayHint>)> {
     // TODO handle relay servers with multiple endpoints better
     let mut relay_hints: Vec<transit::RelayHint> = common_args
@@ -594,11 +581,48 @@ async fn parse_and_connect(
                 .unwrap()],
         )?)
     }
-    let code = code
-        .map(Result::Ok)
-        .or_else(|| (!is_send).then(enter_code))
-        .transpose()?
-        .map(magic_wormhole::Code);
+
+    if code.is_none() && !is_send {
+        code = Some(enter_code()?)
+    }
+
+    // TODO: Apply this change to all usages after an API break
+    // https://github.com/magic-wormhole/magic-wormhole.rs/issues/193
+    // We accept a little breakage in non-interactive use, because this is a security issue
+    // Split the nameplate parsing from the code parsing to ensure we allow non-integer nameplates
+    // until the next breaking release
+    let res: Option<Result<magic_wormhole::Code, _>> = code.as_ref().map(|c| c.parse());
+    let code: Option<magic_wormhole::Code> = match res {
+        Some(Ok(code)) => Some(code),
+        // Check if an interactive terminal is connected
+        Some(Err(
+            err @ (ParseCodeError::SeparatorMissing
+            | ParseCodeError::Password(ParsePasswordError::TooShort { .. })),
+        )) => {
+            // Only fail for the case where the password is < 4 characters.
+            // Anything else will just print an error for now.
+            return Err(err.into());
+        },
+        // If we have an interactive terminal connected, also fail for low entropy
+        Some(Err(err @ ParseCodeError::Password(ParsePasswordError::LittleEntropy { .. })))
+            if std::io::stdin().is_terminal() =>
+        {
+            return Err(err.into())
+        },
+        Some(Err(err)) => {
+            tracing::error!("{} This will fail in the next release.", err);
+            code.map(|c| {
+                let (nameplate, password) = c.split_once("-").unwrap();
+                unsafe {
+                    magic_wormhole::Code::from_components(
+                        magic_wormhole::Nameplate::new_unchecked(nameplate),
+                        magic_wormhole::Password::new_unchecked(password),
+                    )
+                }
+            })
+        },
+        None => None,
+    };
 
     /* We need to track that information for when we generate a QR code */
     let mut uri_rendezvous = None;
@@ -623,10 +647,19 @@ async fn parse_and_connect(
 
             /* Print code and also copy it to clipboard */
             if is_send {
-                if let Some(clipboard) = clipboard {
-                    match clipboard.set_text(mailbox_connection.code().to_string()) {
-                        Ok(()) => log::info!("Code copied to clipboard"),
-                        Err(err) => log::warn!("Failed to copy code to clipboard: {}", err),
+                #[cfg(feature = "clipboard")]
+                {
+                    let clipboard = Clipboard::new()
+                        .map_err(|err| {
+                            tracing::warn!("Failed to initialize clipboard support: {}", err);
+                        })
+                        .ok();
+
+                    if let Some(mut clipboard) = clipboard {
+                        match clipboard.set_text(mailbox_connection.code().to_string()) {
+                            Ok(()) => tracing::info!("Code copied to clipboard"),
+                            Err(err) => tracing::warn!("Failed to copy code to clipboard: {}", err),
+                        }
                     }
                 }
 
@@ -656,7 +689,7 @@ async fn make_send_offer(
             file.display()
         );
     }
-    log::trace!("Making send offer in {files:?}, with name {file_name:?}");
+    tracing::trace!("Making send offer in {files:?}, with name {file_name:?}");
 
     match (files.len(), file_name) {
         (0, _) => unreachable!("Already checked by CLI parser"),
@@ -719,15 +752,6 @@ fn create_progress_handler(pb: ProgressBar) -> impl FnMut(u64, u64) {
     }
 }
 
-fn enter_code() -> eyre::Result<String> {
-    use dialoguer::Input;
-
-    Input::new()
-        .with_prompt("Enter code")
-        .interact_text()
-        .map_err(From::from)
-}
-
 fn print_welcome(term: &mut Term, welcome: Option<&str>) -> eyre::Result<()> {
     if let Some(welcome) = &welcome {
         writeln!(term, "Got welcome from server: {}", welcome)?;
@@ -747,11 +771,17 @@ fn sender_print_code(
         is_leader: false,
     }
     .to_string();
-    writeln!(
-        term,
-        "\nThis wormhole's code is: {} (it has been copied to your clipboard)",
-        style(&code).bold()
-    )?;
+
+    if cfg!(feature = "clipboard") {
+        writeln!(
+            term,
+            "\nThis wormhole's code is: {} (it has been copied to your clipboard)",
+            style(&code).bold()
+        )?;
+    } else {
+        writeln!(term, "\nThis wormhole's code is: {}", style(&code).bold())?;
+    }
+
     writeln!(term, "This is equivalent to the following link: \u{001B}]8;;{}\u{001B}\\{}\u{001B}]8;;\u{001B}\\", &uri, &uri)?;
     let qr =
         qr2term::generate_qr_string(&uri).context("Failed to generate QR code for send link")?;
@@ -776,7 +806,16 @@ fn server_print_code(
     code: &magic_wormhole::Code,
     _: &Option<url::Url>,
 ) -> eyre::Result<()> {
-    writeln!(term, "\nThis wormhole's code is: {}", style(&code).bold())?;
+    if cfg!(feature = "clipboard") {
+        writeln!(
+            term,
+            "\nThis wormhole's code is: {} (it has been copied to your clipboard)",
+            style(&code).bold()
+        )?;
+    } else {
+        writeln!(term, "\nThis wormhole's code is: {}", style(&code).bold())?;
+    }
+
     writeln!(
         term,
         "On the other side, enter that code into a Magic Wormhole client\n"
@@ -826,7 +865,7 @@ async fn send_many(
     transit_abilities: transit::Abilities,
     ctrl_c: impl Fn() -> futures::future::BoxFuture<'static, ()>,
 ) -> eyre::Result<()> {
-    log::warn!("Reminder that you are sending the file to multiple people, and this may reduce the overall security. See the help page for more information.");
+    tracing::warn!("Reminder that you are sending the file to multiple people, and this may reduce the overall security. See the help page for more information.");
 
     /* Progress bar is commented out for now. See the issues about threading/async in
      * the Indicatif repository for more information. Multiple progress bars are not usable
@@ -849,14 +888,14 @@ async fn send_many(
 
     for tries in 0.. {
         if time.elapsed() >= timeout {
-            log::info!(
+            tracing::info!(
                 "{:?} have elapsed, we won't accept any new connections now.",
                 timeout
             );
             break;
         }
         if tries > max_tries {
-            log::info!("Max number of tries reached, we won't accept any new connections now.");
+            tracing::info!("Max number of tries reached, we won't accept any new connections now.");
             break;
         }
 
@@ -907,11 +946,11 @@ async fn send_many(
             match result.await {
                 Ok(_) => {
                     pb.finish();
-                    log::info!("Successfully sent file to someone");
+                    tracing::info!("Successfully sent file to someone");
                 },
                 Err(e) => {
                     pb.abandon();
-                    log::error!("Send failed, {}", e);
+                    tracing::error!("Send failed, {}", e);
                 },
             };
         });
@@ -1144,7 +1183,7 @@ async fn receive_inner_v2(
 }
 
 fn transit_handler(info: TransitInfo) {
-    log::info!("{info}");
+    tracing::info!("{info}");
 }
 
 #[cfg(test)]
